/*
 * Copyright (c) 2004 Iwata <iwata@quasiquote.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <string.h>

#include <gauche/extend.h>
#include "lazy-ffi.h"

static void print_cpointer(ScmObj, ScmPort *, ScmWriteContext *);
SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_CpointerClass, print_cpointer);

static void
print_cpointer(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
	Scm_Printf(out, "#<c-pointer 0x%x>", SCM_C_POINTER(obj)->ptr);
}

ScmObj
Scm_LazyPointerP(ScmObj obj)
{
	if (SCM_C_POINTER_P(obj))
		return SCM_TRUE;
	return SCM_FALSE;

}

ScmObj
Scm_LazyPointerAddress(ScmCpointer *obj)
{
	return Scm_MakeInteger((int)obj->ptr);
}

ScmObj
Scm_LazyMakeNullCpointer()
{
	ScmObj obj;

	ScmCpointer *z = SCM_NEW(ScmCpointer);

	SCM_SET_CLASS(z, SCM_CLASS_CPOINTER);

	z->ptr = NULL;
	return SCM_OBJ(z);
}

ScmObj
Scm_LazyMakeCpointer(int i)
{
	ScmObj obj;

	ScmCpointer *z = SCM_NEW(ScmCpointer);

	SCM_SET_CLASS(z, SCM_CLASS_CPOINTER);

	z->ptr = (void *)i;
	return SCM_OBJ(z);
}

ScmObj
Scm_LazyNullCpointerP(ScmCpointer *obj)
{
	if (obj->ptr == NULL)
		return SCM_TRUE;
	return SCM_FALSE;
}

ScmObj
Scm_LazyAllocateForeignObject(int size)
{
	ScmObj obj;

	ScmCpointer *z = SCM_NEW(ScmCpointer);

	SCM_SET_CLASS(z, SCM_CLASS_CPOINTER);

	z->ptr = malloc(size);
	return SCM_OBJ(z);
}

ScmObj
Scm_LazyAllocateForeignString(int size)
{
	return Scm_LazyAllocateForeignObject(size);
}

ScmObj
Scm_LazyFreeForeignObject(ScmCpointer *obj)
{
	free(obj->ptr);
	return SCM_UNDEFINED;
}

ScmObj
Scm_LazyConvertFromForeignString(ScmCpointer *obj)
{
	return SCM_MAKE_STR((char *)obj->ptr);
}

ScmObj
Scm_LazyConvertToForeignString(ScmString *obj)
{
	ScmObj z;
	char *s;
	size_t len;

	s = Scm_GetString(obj);
	len = strlen(s) + 1;
	z = Scm_LazyAllocateForeignString(len);
	strlcpy(SCM_C_POINTER(z)->ptr, s, len);
	return z;
}

#define Scm_LazyCpointerToXvector(n)					\
	ScmObj								\
	Scm_LazyCpointerTo##n##vector(int size, ScmCpointer *obj)	\
	{								\
		return Scm_Make##n##VectorFromArray(size, obj->ptr);	\
	}

Scm_LazyCpointerToXvector(S8);
Scm_LazyCpointerToXvector(U8);
Scm_LazyCpointerToXvector(S16);
Scm_LazyCpointerToXvector(U16);
Scm_LazyCpointerToXvector(S32);
Scm_LazyCpointerToXvector(U32);
Scm_LazyCpointerToXvector(S64);
Scm_LazyCpointerToXvector(U64);

#define Scm_LazyXvectorToCpointer(n, s)					\
	ScmObj								\
	Scm_Lazy##n##vectorToCpointer(Scm##n##Vector *vec)		\
	{								\
		ScmObj obj;						\
		int size = SCM_##n##VECTOR_SIZE(vec);			\
									\
		obj = Scm_LazyAllocateForeignObject(size);		\
									\
		/* XXX: depend on internal structure */			\
		memcpy(SCM_C_POINTER(obj)->ptr, SCM_##n##VECTOR_ELEMENTS(vec), size * sizeof(s)); \
		return obj;						\
	}

Scm_LazyXvectorToCpointer(S8,  signed char);
Scm_LazyXvectorToCpointer(U8,  unsigned char);
Scm_LazyXvectorToCpointer(S16, short);
Scm_LazyXvectorToCpointer(U16, unsigned short);
Scm_LazyXvectorToCpointer(S32, ScmInt32);
Scm_LazyXvectorToCpointer(U32, ScmUInt32);
Scm_LazyXvectorToCpointer(S64, ScmInt64);
Scm_LazyXvectorToCpointer(U64, ScmUInt64);
