;; Copyright (c) Iwata <iwata@quasiquote.org>
;;
;; Permission is hereby granted, free of charge, to any person obtaining
;; a copy of this software and associated documentation files (the
;; "Software"), to deal in the Software without restriction, including
;; without limitation the rights to use, copy, modify, merge, publish,
;; distribute, sublicense, and/or sell copies of the Software, and to
;; permit persons to whom the Software is furnished to do so, subject to
;; the following conditions:
;;
;; The above copyright notice and this permission notice shall be included
;; in all copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
;; IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;; CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;; TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(define-module lazy-ffi
  (use srfi-11)
  (use gauche.uvector)
  ;; main
  (export lazy-ffi:lazy-ffi?)
  (export lazy-ffi:foreign-pointer-size)
  (export lazy-ffi:strerr)
  (export lazy-ffi:open)
  (export lazy-ffi:close)
  (export lazy-ffi:function-to-pointer)
  (export lazy-ffi:function)
  (export lazy:void lazy:int lazy:double lazy:ptr lazy:char
          lazy:scheme lazy:scheme-ptr lazy:string)
  ;; c-pointer
  (export lazy-ffi:pointer?)
  (export lazy-ffi:pointer-address)
  (export lazy-ffi:make-null-pointer lazy-ffi:make-pointer)
  (export lazy-ffi:null-pointer?)
  (export lazy-ffi:allocate-foreign-object lazy-ffi:allocate-foreign-string)
  (export lazy-ffi:free-foreign-object)
  (export lazy-ffi:convert-from-foreign-string lazy-ffi:convert-to-foreign-string)
  ;; pointer-to-uvector
  (export lazy-ffi:pointer-to-s8vector)
  (export lazy-ffi:pointer-to-u8vector)
  (export lazy-ffi:pointer-to-s16vector)
  (export lazy-ffi:pointer-to-u16vector)
  (export lazy-ffi:pointer-to-s32vector)
  (export lazy-ffi:pointer-to-u32vector)
  (export lazy-ffi:pointer-to-s64vector)
  (export lazy-ffi:pointer-to-u64vector)
  ;; uvector-to-pointer
  (export lazy-ffi:s8vector-to-pointer)
  (export lazy-ffi:u8vector-to-pointer)
  (export lazy-ffi:s16vector-to-pointer)
  (export lazy-ffi:u16vector-to-pointer)
  (export lazy-ffi:s32vector-to-pointer)
  (export lazy-ffi:u32vector-to-pointer)
  (export lazy-ffi:s64vector-to-pointer)
  (export lazy-ffi:u64vector-to-pointer))

(select-module lazy-ffi)
(dynamic-load "lazy-ffi")

(define-constant lazy:void 0)
(define-constant lazy:int 1)
(define-constant lazy:double 2)
(define-constant lazy:ptr 3)
(define-constant lazy:char 4)
(define-constant lazy:scheme 5)
(define-constant lazy:scheme-ptr 6)
(define-constant lazy:string 7)

(define (type-code t)
    (case t
      ((:bool :int) lazy:int)
      ((:void) lazy:void)
      ((:char) lazy:char)
      ((:float :double) lazy:double)
      ((:pointer) lazy:ptr)
      ((:string :symbol) lazy:string)
      ((:scheme-pointer :byte-vector) lazy:scheme-ptr)
      ((:scheme-object) lazy-scheme)
      (else
       (error (format #f "unknown type ~a" t)))))

(define (type-auto obj)
  (cond ((integer? obj) lazy:int)
        ((number? obj)  lazy:double)
        ((char? obj)    lazy:char)
        ((list? obj)    lazy:scheme)
        ((string? obj)  lazy:string)
        ((symbol? obj)  lazy:string)
        ((lazy-ffi:pointer? obj) lazy:ptr)
        (else
         (error (format #f "unknown type ~a" obj)))))


(define (get-opt l)
  (let loop ((l l)
             (return lazy:void)
             (args '())
             (atypelist '()))
    (if (null? l)
        (values return (reverse args) (reverse atypelist))
        ;; keyword?
        (case (car l)
          ((:bool :int :void :char :float :double :pointer :string :symbol :scheme-pointer :byte-vector :scheme-object)
           (loop (cddr l) return (cons (cadr l) args) (cons (type-code (car l)) atypelist)))
          ((:return)
           (loop (cddr l) (type-code (cadr l)) args atypelist))
          (else
           (loop (cdr l) return (cons (car l) args) (cons (type-auto (car l)) atypelist)))))))

;; chicken lazy-ffi like interface
(define (lazy-ffi:function dl fun . l)
  (let-values (((return args atypelist) (get-opt l)))
    (cond ((string? fun)
           (lazy-ffi:call dl return fun args atypelist))
          ((lazy-ffi:pointer? fun)
           (lazy-ffi:call-with-pointer dl return fun args atypelist))
          (else
           (error (format #f "unknown function type ~a" fun))))))

(provide "lazy-ffi")
